public class AutonomesVehikel extends Modell
{
    /**
     * Regelparemeter
     */
    public double PWERT = 0.0008;
    public double SOLLWERT = 50.0;
    
    /**
     * Ergänzungen zu Sensordaten
     */
    public double sensorwert = 0.0; //...wird bei 30cm Entfernung geliefert
    public double distanz = 0.3; //Distanz in Metern, höchstens 0.3m
    public double strahl_x1,strahl_y1,strahl_x2,strahl_y2; //Start- und Endpunkt des Sensorstrahls im Inertialkoordinatensystem        
    //Ende Ergänzungen zu Sensordaten
    
    public boolean DEBUG = false;
    
    private double q = 0.1;
    private double h = 2*q;
    private double D = 1.0;
    private double d = 0.05;
    private double m = 0.3;
    private double J = (1.0/12.0)*m*(d*d+h*h);

    private double x;
    private double y;
    private double phi;
    
    private double vx;
    private double vy;
    private double omega;    
    
    private double[] ex = new double[2];
    private double[] ey = new double[2];

    private double[] FA = new double[2];
    private double[] FB = new double[2];
    
    private double sFA = 0.1;  //Skalar der Antriebskraft links FA
    private double sFB = 0.2;

    private double[][] TT = new double[2][2];
    
    private double[] vekom = new double[3];

    private double[] hilf2 = new double[2];
    private double[] hilf3 = new double[3];
    private double[] hilf31 = new double[3];
    private double[] hilf32 = new double[3];
    private double[] hilf33 = new double[3];
    private double[] hilf34 = new double[3];
    private double[] hilf35 = new double[3];
    private double[] hilf36 = new double[3];
    private double[] hilf37 = new double[3];
    
    private double[] rSA = new double[2];
    private double[] rSB = new double[2];
    
    private double[] va = new double[3];
    private double[] vb = new double[3];
    
    private double[] FAD = new double[3];
    private double[] FBD = new double[3];

    private double LL;      
    private double MM;      
    private double QQ;      
                              
    private double sFZ;
     
    private double[] FZ = new double[2];

    private double[] EULER = new double[3];
    
    private void set(double[] vektor,double x, double y)
    {
        vektor[0]=x;
        vektor[1]=y;
    }
    private void set(double[] vektor,double x, double y, double z)
    {
        vektor[0]=x;
        vektor[1]=y;
        vektor[2]=z;
    }

    private void dreh(double[] erg, double[][] TT,double[] vektor)
    {
        erg[0] = TT[0][0]*vektor[0]+TT[0][1]*vektor[1];
        erg[1] = TT[1][0]*vektor[0]+TT[1][1]*vektor[1];
    }

    private void cross(double[] erg, double[] x,double[] y)
    {
        erg[0] = x[1]*y[2]-x[2]*y[1];
        erg[1] = x[2]*y[0]-x[0]*y[2];
        erg[2] = x[0]*y[1]-x[1]*y[0];
    }

    private void sum(double[] erg, double[] x, double[] y)
    {
        for(int i=0;i<erg.length;i++)
            erg[i]=x[i]+y[i];
    }

    private void skalar(double[] erg, double[] x, double skalar)
    {
        for(int i=0;i<erg.length;i++)
            erg[i]=x[i]*skalar;        
    }
    
    private void show(double[] vektor, String name)
    {
        if(!DEBUG)
            return;
        System.out.print(name+"=");
        for(int i=0;i<vektor.length;i++)
        {
            System.out.print(vektor[i]+" ");            
        }
        System.out.println();
    }
    private void show(double skalar, String name)
    {
        if(DEBUG)
            System.out.println(name+"="+skalar);
    }
    
    public AutonomesVehikel()
    {
        f = new double[6];
    }

    public int holeAnzahlGleichungen()
    {
        return 6;
    }    

    public boolean rundeVollendet(double x, double y, double phi)
    {
        if(y<0.0 && x<-0.001 && x>-0.03)
        {
            return true;
        }
        else
        {
            return false;
        }
    }
    
    public boolean kollision(double x, double y, double phi)
    {
        Par.setXYPHI(x,y,phi);
        
        double x1 = Par.stransX(0.5*Par.d,Par.q);        
        double y1 = Par.stransY(0.5*Par.d,Par.q);        
        
        double x2 = Par.stransX(-0.5*Par.d,Par.q);        
        double y2 = Par.stransY(-0.5*Par.d,Par.q);        
       
        double x3 = Par.stransX(-0.5*Par.d,-Par.q);        
        double y3 = Par.stransY(-0.5*Par.d,-Par.q);        
        
        double x4 = Par.stransX(0.5*Par.d,-Par.q);        
        double y4 = Par.stransY(0.5*Par.d,-Par.q);        
        
        if(x1>-0.5*Par.BREITE_HINDERNIS && x1<0.5*Par.BREITE_HINDERNIS 
        && y1>-0.5*Par.HOEHE_HINDERNIS && y1<0.5*Par.HOEHE_HINDERNIS)
        {
            return true;
        }
        if(x2>-0.5*Par.BREITE_HINDERNIS && x2<0.5*Par.BREITE_HINDERNIS 
        && y2>-0.5*Par.HOEHE_HINDERNIS && y2<0.5*Par.HOEHE_HINDERNIS)
        {
            return true;
        }
        if(x3>-0.5*Par.BREITE_HINDERNIS && x3<0.5*Par.BREITE_HINDERNIS 
        && y3>-0.5*Par.HOEHE_HINDERNIS && y3<0.5*Par.HOEHE_HINDERNIS)
        {
            return true;
        }
        if(x4>-0.5*Par.BREITE_HINDERNIS && x4<0.5*Par.BREITE_HINDERNIS 
        && y4>-0.5*Par.HOEHE_HINDERNIS && y4<0.5*Par.HOEHE_HINDERNIS)
        {
            return true;
        }

        return false;
    }

    
    
    public void aktualisiereSensorwert(double x, double y, double phi)
    {
//    public double sensorwert = 0.0; //...wird bei 30cm Entfernung geliefert
//    public double strahl_x1,strahl_y1,strahl_x2,strahl_y2; //Start- und Endpunkt des Sensorstrahls im Inertialkoordinatensystem        
      //Aufpunkt für Sensor:
      double x1 = x;
      double y1 = y;
      //zweiter Punkt für Sensor im Körperkoordinatensystem:
      double xk2 = 1.0/Math.sqrt(2.0);
      double yk2 = 1.0/Math.sqrt(2.0);
      //...transformiert:
      double x2 = x + xk2*Math.cos(phi) - yk2*Math.sin(phi);
      double y2 = y + xk2*Math.sin(phi) + yk2*Math.cos(phi);
      
      //Richtungsvektor:
      double rx = x2-x1;
      double ry = y2-y1;
      
      double faktor = 0.0; //in Millimeterschritten
      
      for(int i=0;i<300.0;i++) //bis 300mm
      {
          faktor+=0.001;
          //Prüfen, ob Geradenpunkt in Hindernis:
          double xs = x1+faktor*rx;
          double ys = y1+faktor*ry;
          
          if(xs>=-Par.BREITE_HINDERNIS*0.5 && xs<=Par.BREITE_HINDERNIS*0.5
          && ys>=-Par.HOEHE_HINDERNIS*0.5 && ys<=Par.HOEHE_HINDERNIS*0.5)
          {
              break;
          }
      }
      distanz = faktor;
      sensorwert = (0.3-faktor)*255.0/0.3;
      strahl_x1 = x;
      strahl_y1 = y;
      strahl_x2 = x1+faktor*rx;
      strahl_y2 = y1+faktor*ry;       
      
      //Neue Antriebskraft berechnen:
//      sFA = 0.1 - 0.0005*(50.0-sensorwert);
//      sFB = 0.1 + 0.0005*(50.0-sensorwert);
      sFA = 0.1 - PWERT*(SOLLWERT-sensorwert);
      sFB = 0.1 + PWERT*(SOLLWERT-sensorwert);
    }
    
    public void berechneRechteSeite(double[] yy, double t)
    {
        
        if(DEBUG)
            System.out.println("*** RECHTE SEITE ***");        
        x = yy[0];
        y = yy[1];
        phi = yy[2];
    
        vx    = yy[3];
        vy    = yy[4];
        omega = yy[5];
        
        if(DEBUG) System.out.println("x="+x);        
        if(DEBUG) System.out.println("y="+y);        
        if(DEBUG) System.out.println("phi="+phi);        
        if(DEBUG) System.out.println("vx="+vx);        
        if(DEBUG) System.out.println("vy="+vy);        
        if(DEBUG) System.out.println("omega="+omega);        
             
        ex[0] = Math.cos(phi);
        ex[1] = Math.sin(phi);
        ey[0] = -Math.sin(phi);
        ey[1] = Math.cos(phi);
        FA[0] = sFA*ex[0];
        FA[1] = sFA*ex[1];
        FB[0] = sFB*ex[0];
        FB[1] = sFB*ex[1];
    
        //Drehmatrix:    
        //TT = [ex,ey];
        TT[0][0] = ex[0];
        TT[1][0] = ex[1];
        TT[0][1] = ey[0];
        TT[1][1] = ey[1];
        
        //Geschwindigkeiten in A und B:
        //vekom = [0;0;omega];
        set(vekom,0.0,0.0,omega);        
        show(omega,"omega");
        show(vekom,"vekom");
        
        //rSA = TT*[0;q];
        set(hilf2,0.0,q);
        dreh(rSA,TT,hilf2);
        
        //rSB = -rSA;
        set(rSB,-rSA[0],-rSA[1]);

        
        show(rSA,"rSA");
        show(rSB,"rSB");
        
        //va = [vx;vy;0] + cross(vekom,[rSA(1);rSA(2);0]);        
        set(hilf32,rSA[0],rSA[1],0.0);
        show(hilf32,"hilf32");
        cross(hilf3,vekom,hilf32);
        show(hilf3,"hilf3");
        
        
        set(hilf33,vx,vy,0.0);
        sum(va,hilf33,hilf3);
        
        //vb = [vx;vy;0] + cross(vekom,[rSB(1);rSB(2);0]);        
        set(hilf32,rSB[0],rSB[1],0.0);
        show(hilf3,"hilf32");
        cross(hilf3,vekom,hilf32);
        show(hilf3,"hilf3");
        set(hilf33,vx,vy,0.0);
        sum(vb,hilf33,hilf3);

        show(va,"va");
        show(vb,"vb");
        
        //FAD = -D*va;
        skalar(FAD,va,-D);
        
        //FBD = -D*vb;
        skalar(FBD,vb,-D);

        show(FAD,"FAD");
        show(FBD,"FBD");
        
        LL = FA[0]+FAD[0]+FB[0]+FBD[0];      
        MM = FA[1]+FAD[1]+FB[1]+FBD[1];      
        show(LL,"LL");
        show(MM,"MM");
        
        QQ = vx*omega*Math.cos(phi) + vy*omega*Math.sin(phi);
        
        sFZ = LL*Math.sin(phi) - MM*Math.cos(phi) + QQ*m;
     
        //FZ = sFZ*ey;
        skalar(FZ,ey,sFZ);
        
                                    
        f[0] = vx;
        f[1] = vy;
        f[2] = omega;
    
        f[3] = (LL+FZ[0])/m;
        f[4] = (MM+FZ[1])/m;

        //EULER = cross([rSA(1);rSA(2);0],[FA(1)+FAD(1);FA(2)+FAD(2);0]) + cross([rSB(1);rSB(2);0],[FB(1)+FBD(1);FB(2)+FBD(2);0]);
        set(hilf31,rSA[0],rSA[1],0.0);
        set(hilf32,FA[0]+FAD[0],FA[1]+FAD[1],0.0);
        
        set(hilf33,rSB[0],rSB[1],0.0);
        set(hilf34,FB[0]+FBD[0],FB[1]+FBD[1],0.0);
    
        cross(hilf35,hilf31,hilf32);
        cross(hilf36,hilf33,hilf34);
        
        sum(EULER,hilf35,hilf36);
        
        f[5] = EULER[2]/J;   
    }
}
