
/**
 * Nützliche mathematische Methoden
 */
public class Hilfsmethoden
{
    /**
     * Liefert den Schnittpunkt zweier Geraden oder null, falls diese fast parallel laufen.
     * Die Geraden werden über jeweils zwei Punkte in der Ebene beschrieben.
     * Gerade1: x1,y1,x2,y2
     * Gerade2: x3,y3,x4,y4
     */
    public static double[] schnittpunkt(double x1,double y1, double x2, double y2,
                                        double x3, double y3, double x4, double y4)
    {
        double r1x = x2-x1;
        double r1y = y2-y1;
        double r2x = x4-x3;
        double r2y = y4-y3;
        
        double s1 = Math.sqrt(r1x*r1x+r1y*r1y);
        double s2 = Math.sqrt(r2x*r2x+r2y*r2y);
        
        r1x/=s1;
        r1y/=s1;
        r2x/=s2;
        r2y/=s2;
        
        double kreuzprod = r1x*r2y-r1y*r2x;
        
        //Auf Parallelität prüfen:
        if(kreuzprod>-0.0001 && kreuzprod<0.0001)
            return null;
        
        //Lineares Gleichungssystem bilden:
        //P1+a*r1=P3+b*r2
        //[r1,-r2]*[a;b]=P3-P1
        //A       *c    =b
        double[] ab = cramer(r1x,r2x,r1y,r2y,x3-x1,y3-y1);
        
        return new double[] {x1+ab[0]*r1x,y1+ab[0]*r1y};
    }

    /**
     * liefert den Schnittpunkt und die Faktoren a,b der Geraden.
     */
    public static double[] schnittpunkt2(double x1,double y1, double x2, double y2,
                                        double x3, double y3, double x4, double y4)
    {
        double r1x = x2-x1;
        double r1y = y2-y1;
        double r2x = x4-x3;
        double r2y = y4-y3;
        
        double s1 = Math.sqrt(r1x*r1x+r1y*r1y);
        double s2 = Math.sqrt(r2x*r2x+r2y*r2y);
        
        r1x/=s1;
        r1y/=s1;
        r2x/=s2;
        r2y/=s2;
        
        double kreuzprod = r1x*r2y-r1y*r2x;
        
        //Auf Parallelität prüfen:
        if(kreuzprod>-0.0001 && kreuzprod<0.0001)
            return null;
        
        //Lineares Gleichungssystem bilden:
        //P1+a*r1=P3+b*r2
        //[r1,-r2]*[a;b]=P3-P1
        //A       *c    =b
        double[] ab = cramer(r1x,r2x,r1y,r2y,x3-x1,y3-y1);
        
        return new double[] {x1+ab[0]*r1x,y1+ab[0]*r1y,ab[0],ab[1]};
    }
    
    /**
     * A=[a11,a12;a21,a22] b=[b1;b2]
     * Ax=b
     * Liefert Vektor x, keine Kontrolle, ob lösbar!
     */
    public static double[] cramer(double a11,double a12, double a21, double a22, double b1, double b2)
    {
        double det = a11*a22-a12*a21;
        double detX = b1*a22-b2*a21;
        double detY = a11*b2-a12*b1;
        
        return new double[]{detX/det,detY/det};
    }
    
    public static void main(String[] args)
    {
        //Tests:
        // 0   1    4   4
        //     
        // 0   0,   0   1
        //Schnittpunkt: 4,0
        double[] erg = schnittpunkt(0.0,0.0,1.0,0.0,  4.0,0.0,4.0,1.0);
        System.out.println("Schnittpunkt: x="+erg[0]+" y="+erg[1]);
        //Tests:
        // 0   1    0   4
        //     
        // 0   1,   4   0
        //Schnittpunkt: 2,2
        double[] erg2 = schnittpunkt(0.0,0.0,1.0,1.0,  0.0,4.0,4.0,0.0);
        System.out.println("Schnittpunkt: x="+erg2[0]+" y="+erg2[1]);
    }
}
